package serial
// Serial io as raw as possible. 8 bits, no parity 1 stop bit, channel to provide DTR control
//
// comName := "com3" 
// baudRate := 57600
// debug := false
//
// chanFromSerial, chanToSerial, charDTRSerial, chanQuitSerial, err := SerialChannels( comName, baud, debug)
//
// chanFromSerial = byte data channel from serial
// chanToSerial - byte data channel to serial 
// chanDTRSerial - bool channel to control DTR - logical 1 means DTR active - rs232 level - +3-+15 volts 
// chanQuitSerial - send a bit on this channel to terminate
//
// based on http://code.google.com/p/goserial/
//
// Works with go version 1, tested on Windows 7(64 bit) and xp (32bit)
import (
	"errors"
	"fmt"
	"os"
	"sync"
	"syscall"
	"time"
	"unsafe"
)

type serialPort struct {
	fp  *os.File
	fd syscall.Handle
	rl sync.Mutex
	wl sync.Mutex
	ro *syscall.Overlapped
	wo *syscall.Overlapped
}

type structDCB struct {
	DCBlength, BaudRate uint32
	flags [4]byte
	wReserved, XonLim, XoffLim uint16
	ByteSize, Parity, StopBits byte
	XonChar, XoffChar, ErrorChar, EofChar, EvtChar byte
	wReserved1 uint16
}

type structTimeouts struct {
	ReadIntervalTimeout uint32
	ReadTotalTimeoutMultiplier uint32
	ReadTotalTimeoutConstant uint32
	WriteTotalTimeoutMultiplier uint32
	WriteTotalTimeoutConstant uint32
}

func SerialChannels( name string, baud int, conLog chan string, debug bool) (chan byte, chan byte, chan bool, chan bool, chan bool, error) {
	p := new(serialPort)
	var err error

	chFromSerial := make(chan byte,8192)
	chToSerial := make(chan byte,8192)
	chQuitSerial := make(chan bool)
	chDTRSerial := make(chan bool)
	chQuitSerial1 := make(chan bool) //**********PBP
	chQuitSerial2 := make(chan bool) //**********PBP
	
	if len(name) >0 && name[0] != '\\' {
		name = "\\\\.\\" + name
	}
	p.fd, err = syscall.CreateFile( syscall.StringToUTF16Ptr(name),
		syscall.GENERIC_READ|syscall.GENERIC_WRITE, 0, nil, syscall.OPEN_EXISTING,
		syscall.FILE_ATTRIBUTE_NORMAL|syscall.FILE_FLAG_OVERLAPPED, 0)
	if err == nil {
		p.fp = os.NewFile( uintptr(p.fd), name)
	}
//	defer p.fp.Close() //--------------------------------

	if err == nil {
		err = setCommState(p.fd, baud)
	}
	if err == nil {
		err = setupComm(p.fd, 64, 64)
	}
	if err == nil {
		err = setCommTimeouts(p.fd)
	}
	if err == nil {
		err = setCommMask(p.fd)
	}
	if err == nil {
		p.ro, err = newOverlapped()
	}
	if err == nil {
		p.wo, err = newOverlapped()
	}
	if  err != nil {
		err = errors.New( "OpenPort ERROR: " + err.Error())
	} else {
		chFromSerial = make( chan byte, 8192)
		chQuitFrom := make( chan bool)
		go func() {
			var count int
			var done uint32
			then := time.Now()
			datain := make( []byte, 8192)
			for q := false; q == false; {

//func (p *serialPort)read(datain []byte, done uint32) (count int) {
			count = p.read(datain, done)
/*
//**********************************************************************
				pb.rl.Lock()
				//defer pb.rl.Unlock()
				readerr := resetEvent(pb.ro.HEvent)
				if readerr == nil {
					readerr = syscall.ReadFile(pb.fd , datain, &done, pb.ro)
					time.Sleep( 1e3) //*****PBP
					if readerr == syscall.ERROR_IO_PENDING {
						count, readerr = getOverlappedResult(pb.fd, pb.ro)
					} else {
						count = int(done)
					}
				}
				pb.rl.Unlock()
//**********************************************************************
*/				
				din := datain[:count]

				if debug && len(din) > 0 {
					now := time.Now()
					ds1 := fmt.Sprintf("(%v(%d)", now.Sub(then), len(din))
					then = now
					for _, c := range( din) {
						if c < byte(0x20) || c > byte(0x7E) {
							ds1 += fmt.Sprintf("{%02X}",c)
						} else {
							ds1 += fmt.Sprintf("%c", c)
						}
					}
					conLog <- " in-<"+ds1+")\n"
				}
				for _, c := range( din) {
					chFromSerial <- c
				}
				select {
  				case q = <-chQuitFrom:
				default:
					time.Sleep(10)
				}
			}
			chQuitSerial1 <- true //**********PBP
		}()

		chToSerial = make( chan byte, 8192)
		chQuitTo := make( chan bool)
		go func() {
			dataout := make( []byte, 128) //*256*512*128************PBP
			var count, wcount int
			var readflag bool
			then := time.Now()
			//counta := 0
			for q := false; q == false; {
				select {
				case q = <- chQuitTo:
				case dataout[ 0] = <- chToSerial:
					for readflag, wcount = true, 1; readflag && wcount < 128; wcount++ { //*256*512*128**PBP
						select {
						case dataout[wcount] = <- chToSerial:
						default:
							readflag = false
							wcount--
						}
					}

					dout := dataout[:wcount]
					
					if debug {
						now := time.Now()
						ds1 := fmt.Sprintf("\t[%v[%d]", now.Sub(then), len(dout))
						then = now
						for _, c := range( dout) {
							if c < byte(0x20) || c > byte(0x7E) {
								ds1 += fmt.Sprintf("{%02X}",c)
							} else {
								ds1 += fmt.Sprintf("%c", c)
							}
						}
						conLog <- ds1+"]\n"
					}
					
//**********************************************************************

					p.wl.Lock()
					//defer p.wl.Unlock()
					writeerr := resetEvent(p.wo.HEvent)
					if  writeerr == nil {
						var n uint32 = uint32( wcount)
						writeerr = syscall.WriteFile( p.fd, dout, &n, p.wo)
						count = int( n)
						time.Sleep( 1e3) //*1e5*1e4*1e3*1e2**
						if writeerr == syscall.ERROR_IO_PENDING {
							//fmt.Println(counta," *** writeerr ***",writeerr)
							//counta++
							for i := 0; i < 1000; i++ {
								time.Sleep( 1e3)
								count, writeerr = getOverlappedResult(p.fd, p.wo)
								if count != wcount {
									conLog <- fmt.Sprintf( "Serial port write erro3 %d %s %v %v %v\n", i,  writeerr, writeerr, count, wcount)
									time.Sleep( 1e9)
								} else {
									break
								}
								////time.Sleep( 1e5)
							}
							//fmt.Println("--- writeerr ---",writeerr)
						}
					} else {
						fmt.Println("\n*** not nill ***\n",writeerr)
					}
					p.wl.Unlock()
					//fmt.Println(counta," *** wl.Unlock() ***",writeerr)
					if count != wcount {
						conLog <- fmt.Sprintf( "Serial port write error1 %s %v %v %v\n", writeerr, writeerr, count, wcount)
					}

//**********************************************************************
						
				}
			}
			chQuitSerial1 <- true //**********PBP
		}()

		chQuitDTR := make( chan bool)
		go func() {
			for q := false; q == false; {
				select {
				case q = <- chQuitDTR:
				case dtr := <-chDTRSerial:
					if dtr {
						escapeCommFunction( p.fd, 5)
					} else {
						escapeCommFunction( p.fd, 6)
					}
				}
			}
			chQuitSerial1 <- true //**********PBP
		}()
		
		go func() {
			<- chQuitSerial
			fmt.Println("**** from chanQuitSerial ****")
			err =syscall.GetLastError()
			fmt.Println("**** GetLastError()",err,"****")
			err = syscall.CloseHandle(p.fd)
			fmt.Println("**** CloseHandle()",err,"****")
			p.fp.Close() //-------------------------------
			chQuitFrom <- true
			time.Sleep(1e8)
			<- chQuitSerial1 //**********PBP
			fmt.Println("**** from chQuitFrom ****")
			chQuitTo <- true
			time.Sleep(1e8)
			<- chQuitSerial1 //**********PBP
			fmt.Println("**** from chQuitTo ****")
			chQuitDTR <- true
			time.Sleep(1e8)
			<- chQuitSerial1 //**********PBP
			fmt.Println("\n**** close down serial ****\n")
			time.Sleep(2e9)
			chQuitSerial2 <- true //**********PBP
		}()
	}

	return chFromSerial, chToSerial, chDTRSerial, chQuitSerial, chQuitSerial2, err //**********PBP
}

//**********************************************************************
//func (p *serialPort) Read(buf []byte) (int, error) {
func (p *serialPort)read(datain []byte, done uint32) (count int) {
	p.rl.Lock()
	defer p.rl.Unlock()
	readerr := resetEvent(p.ro.HEvent)
	if readerr == nil {
		readerr = syscall.ReadFile(p.fd , datain, &done, p.ro)
		time.Sleep( 1e3) //*****PBP
		if readerr == syscall.ERROR_IO_PENDING {
			count, readerr = getOverlappedResult(p.fd, p.ro)
		} else {
			count = int(done)
		}
	}
	//rl.Unlock()
	return count
}

//**********************************************************************

//**********************************************************************
//**********************************************************************

var (
	nSetCommState,
	nEscapeCommFunction,
	nSetCommTimeouts,
	nSetCommMask,
	nSetupComm,
	nGetOverlappedResult,
	nCreateEvent,
	nResetEvent uintptr
)

func init() {
	handle, err := syscall.LoadLibrary("kernel32.dll")
	if err == nil {
		nSetCommState = getProcAddr(handle, "SetCommState")
		nEscapeCommFunction = getProcAddr(handle, "EscapeCommFunction")
		nSetCommTimeouts = getProcAddr(handle, "SetCommTimeouts")
		nSetCommMask = getProcAddr(handle, "SetCommMask")
		nSetupComm = getProcAddr(handle, "SetupComm")
		nGetOverlappedResult = getProcAddr(handle, "GetOverlappedResult")
		nCreateEvent = getProcAddr(handle, "CreateEventW")
		nResetEvent = getProcAddr(handle, "ResetEvent")
		syscall.FreeLibrary(handle)
	} else {
		panic("LoadLibrary " + err.Error())
	}
}

func getProcAddr(handle syscall.Handle, name string) uintptr {
	addr, err := syscall.GetProcAddress(handle, name)
	if err != nil {
		panic(name + " " + err.Error())
	}
	return addr
}

func escapeCommFunction(handle syscall.Handle, fparam int) error {
	r, _, err := syscall.Syscall(nEscapeCommFunction, 2, uintptr(handle), uintptr(fparam), 0)
	if r == 0 {
		return err
	}
	return nil
}

func setCommState(handle syscall.Handle, baud int) error {
	var params structDCB
	params.DCBlength = uint32(unsafe.Sizeof(params))
	params.flags[0] = 0x01 // fBinary
	params.BaudRate = uint32(baud)
	params.ByteSize = 8
	r, _, err := syscall.Syscall(nSetCommState, 2, uintptr(handle), uintptr(unsafe.Pointer(&params)), 0)
	if r == 0 {
		return err
	}
	return nil
}
//
// timeout set to be as short as possible
//
func setCommTimeouts(handle syscall.Handle) error {
	var timeouts structTimeouts
	const MAXDWORD = 1<<32 - 1
//	const MAXDWORD = 1<<16 - 1 //******PBP
	timeouts.ReadIntervalTimeout = MAXDWORD //******PBP
	timeouts.ReadTotalTimeoutMultiplier = 0 //0 ******PBP
	timeouts.ReadTotalTimeoutConstant =  0 //0 ******PBP
	timeouts.WriteTotalTimeoutMultiplier = MAXDWORD //1 ******PBP
	timeouts.WriteTotalTimeoutConstant = 0 //1 ******PBP
	r, _, err := syscall.Syscall(nSetCommTimeouts, 2, uintptr(handle), uintptr(unsafe.Pointer(&timeouts)), 0)
	if r == 0 {
		return err
	}
	return nil
}

func setupComm(handle syscall.Handle, in, out int) error {
	r, _, err := syscall.Syscall(nSetupComm, 3, uintptr(handle), uintptr(in), uintptr(out))
	if r == 0 {
		return err
	}
	return nil
}

func setCommMask(handle syscall.Handle) error {
	const EV_RXCHAR = 0x0001
	r, _, err := syscall.Syscall(nSetCommMask, 2, uintptr(handle), EV_RXCHAR, 0)
	if r == 0 {
		return err
	}
	return nil
}

func resetEvent(handle syscall.Handle) error {
	r, _, err := syscall.Syscall(nResetEvent, 1, uintptr(handle), 0, 0)
	if r == 0 {
		return err
	}
	return nil
}

func newOverlapped() (*syscall.Overlapped, error) {
	var overlapped syscall.Overlapped
	r, _, err := syscall.Syscall6(nCreateEvent, 4, 0, 1, 0, 0, 0, 0)
	if r == 0 {
		return nil, err
	}
	overlapped.HEvent = syscall.Handle(r)
	return &overlapped, nil
}

func getOverlappedResult(handle syscall.Handle, overlapped *syscall.Overlapped) (int, error) {
	var n int
	r, _, err := syscall.Syscall6(nGetOverlappedResult, 4,
	uintptr(handle),
	uintptr(unsafe.Pointer(overlapped)),
	uintptr(unsafe.Pointer(&n)), 1, 0, 0)
	if r == 0 {
		return n, err
	}
	return n, nil
}
